/***********************************************************
Evolutionary Algorithm code utilizing Genetic Algorithm

William Westrick
Nicole Rodia
************************************************************/

#include <avr/io.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "main.h"
#include "evolution.h"
#include "uart.h"
#include "mouse_sensor.h"

/***********************************************************
* void initializePopulation(void)
* 		Sets up the initial population with random vectors
*			and a bogus (but large) fitness
************************************************************/
void initializePopulation(void)
{
	//randomize all the phenotypes
	unsigned char i;
	for (i=0;i<NUM_PHENOTYPES;i++)
	{
		//generate random number between -100 and +100
		population[i].motors.chars[0] = (char)((float)rand()*200/RAND_MAX)-100;
		population[i].motors.chars[1] = (char)((float)rand()*200/RAND_MAX)-100;
		population[i].motors.chars[2] = (char)((float)rand()*200/RAND_MAX)-100;
		population[i].fitness = 10000; //fitness goes to 0 when perfect
	}
}

/***********************************************************
* void runPhenotype(phenotype individual)
* 		Sets the motors to the requested phenotype values
*			and sets the motor countdown timer
************************************************************/
void runPhenotype(phenotype individual)
{
	//set the motor outputs appropriately
	SetOutput(0,individual.motors.chars[0]);
	SetOutput(1,individual.motors.chars[1]);
	SetOutput(2,individual.motors.chars[2]);
	motorTime = motorTimeMax;
}

/***********************************************************
* void stopPhenotype(void)
* 		Turns the motors off, and flags completion
************************************************************/
void stopPhenotype(void)
{
	//stop the motors
	SetOutput(0,0);
	SetOutput(1,0);
	SetOutput(2,0);
	//flag that the phenotype is done
	doneRunningMember=1;
}

/***********************************************************
* void ClearResult(void)
* 		Clears the accumulated position information
************************************************************/
void ClearResult(void)
{
	//clear the position information
	xval=0;
	xval2=0;
	yval=0;
	omega=0;
}

/***********************************************************
* void CalculateResult(void)
* 		Finds the result of the motion
************************************************************/
void CalculateResult(void)
{
	//grab the motion information from the mouse sensors
	Position[0] = xval2;
	Position[1] = yval;
	Position[2] = omega;
	
	//FOR SIMULATION:
	// this simulates the omniwheel motion for given motor values
	//Position[0] = (long)(C*motoroutputs[0].totalvalue + D*motoroutputs[1].totalvalue + E*motoroutputs[2].totalvalue);
	//Position[1] = (long)(A*motoroutputs[0].totalvalue + B*motoroutputs[1].totalvalue);
	//Position[2] = (long)(F*motoroutputs[0].totalvalue + G*motoroutputs[1].totalvalue + H*motoroutputs[2].totalvalue);
}

/***********************************************************
* long DetermineFitness(long, long, long)
* 		This calculates the fitness of the last solution
************************************************************/
unsigned long DetermineFitness(signed long DesiredX,signed long DesiredY,signed long DesiredTheta)
{
	//calculate a fitness value
	//use the distance from the desired position as the fitness
	// zero fitness is ideal, that means it is in the right spot
	unsigned long measured_fitness= 0;
	measured_fitness += abs(DesiredX-Position[0]);
	measured_fitness += abs(DesiredY-Position[1]);
	measured_fitness += abs(DesiredTheta-Position[2]);
	return measured_fitness;
}

/***********************************************************
* phenotype BreedPhenotype(phenotype pheno)
* 		Cross-breed the phenotypes and possibly mutate
************************************************************/
phenotype BreedPhenotype(phenotype pheno)
{
	unsigned char k;
	unsigned char parent;
	unsigned char crossoverpoint;
	unsigned long parentcontribution[2];

	//temporary phenotype
	phenotype newPheno=pheno;

	//randomly select a parent
	parent = NUM_PHENOTYPES*(float)rand()/0x7FFF;

	//pick a random crossover point
	crossoverpoint = ((float)rand()*((3*8)-1)/RAND_MAX)+1;
	//build bitmasks for each of the parents for crossover
	parentcontribution[0]=0;
	parentcontribution[1]=0;
	for (k=0;k<crossoverpoint;k++)
	{
		parentcontribution[0] |= (1<<k);
	}
	for (k=(3*8-1);k>=crossoverpoint;k--)
	{
		parentcontribution[1] |= (1<<k);
	}

	//crossover the bits
	if ((float)rand()<(RAND_MAX/2))
	{
		newPheno.motors.value = (parentcontribution[0] & pheno.motors.value) |
								(parentcontribution[1] & population[parent].motors.value);
	} else {
		newPheno.motors.value = (parentcontribution[0] & pheno.motors.value) |
								(parentcontribution[1] & population[parent].motors.value);
	}
	//mutate some bits randomly
	if (rand() < MUTATION_RATE)
	{
		newPheno = mutatePhenotype(newPheno);
	}

	//check for zero'd solutions and flip a few bits in them to keep it from propagating
	if ((newPheno.motors.value&(~0x00808080)) == 0)
	{
		newPheno.motors.value ^= ((1<<(char)(3*8*(float)rand()/RAND_MAX))&(~0x00808080));
		newPheno.motors.value ^= ((1<<(char)(3*8*(float)rand()/RAND_MAX))&(~0x00808080));
		newPheno.motors.value ^= ((1<<(char)(3*8*(float)rand()/RAND_MAX))&(~0x00808080));
		newPheno.motors.value ^= ((1<<(char)(3*8*(float)rand()/RAND_MAX))&(~0x00808080));
	}

	//return the new phenotype
	return newPheno;
}

/***********************************************************
* phenotype mutatePhenotype(phenotype pheno) 
* 		Randomly flip up to 4 bits within the phenotype
*		This allows the algorithm to develop new solutions
*		 and tries to keep the population
*		 from becoming homogenous.
************************************************************/
phenotype mutatePhenotype(phenotype pheno) 
{
	//store the input phenotype into a temporary variable
	phenotype newPheno=pheno;

	//mutate(flip) random bits
	//flip a random (non-sign) bit within a motor char
	newPheno.motors.value ^= ((1<<((char)(7*(float)rand()/RAND_MAX)))&(~0x00000080));	
	//flip a random (non-sign) bit within a motor char
	newPheno.motors.value ^= ((1<<((char)(7*(float)rand()/RAND_MAX)+8))&(~0x00008000)); 
	//flip a random (non-sign) bit within a motor char	
	newPheno.motors.value ^= ((1<<((char)(7*(float)rand()/RAND_MAX)+16))&(~0x00800000)); 

	//flip a random sign bit
	newPheno.motors.value ^= ((1<<(char)(float)rand()*3*8/RAND_MAX)&0x00808080);

	//return the new phenotype
	return newPheno;
}

/***********************************************************
* void initializeEvolution(void)
* 		Sets up the initial conditions and runs initialization
************************************************************/
void initializeEvolution(void)
{
	//set the ranomize seed
	srand(-1189);
	//clear the status variables
	avgfitness=0;
	minfitness=0;
	epoch=0;
	hasInitialized=0;
	//initialize the population
	initializePopulation();
}

/***********************************************************
* void StartNewEpoch(void)
* 		Sets up a new epoch, clears statistics,
*		 and also checks for homogenous populations
************************************************************/
void StartNewEpoch(void)
{
	unsigned char k;
	epoch++;

	//ensure that an entirely homogenous population has not developed
	if ((displayavgfitness == minfitness) && (displayavgfitness == WorstFitness))
	{
		if (displayavgfitness>FITNESS_THRESHOLD)
		{
			//all the phenotypes are the same fitness, and they're all bad, mutate a bunch of them randomly
			for (k=0;k<((float)rand()*NUM_PHENOTYPES/RAND_MAX);k++)
			{
				population[k] = mutatePhenotype(population[k]);
				population[k].fitness = 100000;
			}
		}
	}

	//setup cycle for next generation
	curMember=0;
	avgfitness=0;
	WorstFitness = 0;
	runningMember = 0;
	minfitness = population[0].fitness;
}

/***********************************************************
* void StartNextMember(void)
* 		Sets up conditions for running the next phenotype
*		 if the net one is already an accepted solution,
*		  it is not changed, but still runs
************************************************************/
void StartNextMember(void) 
{
	//set the status flags
	runningMember=1;
	doneRunningMember=0;
	//clear the position information
	ClearResult();
	if (hasInitialized)
	{
		//done initial test of population, mutate!
		if (population[curMember].fitness>FITNESS_THRESHOLD)
		{
			//breed and mutate the phenotype
			newPhenotype = BreedPhenotype(population[curMember]);
		} else {
			//don't change the phenotype, but still run it
			PORTD &= ~(1<<PD2); //turn on led to indicate this is an accepted solution 
			newPhenotype = population[curMember];
		}
	} else {
		//just use the same phenotype
		newPhenotype = population[curMember];
	}
	//run the phenotype
	runPhenotype(newPhenotype);
}

/***********************************************************
* void UpdateMember(void)
* 		Reads the surface quality from the mouse sensors
************************************************************/
void UpdateMember(void) 
{
	//see what happened
	CalculateResult();
	//find the fitness
	newPhenotype.fitness=DetermineFitness(0,6000,0);
	
	//initial run?
	if (hasInitialized)
	{
		if (   (newPhenotype.fitness < population[curMember].fitness)
			|| (population[curMember].fitness<=FITNESS_THRESHOLD))
		{
			//store the new phenotype if it is better than before
			// also, update the fitness if it was previously deemed acceptable
			// this ensures that a faulty reading does not cause a bad solution to live forever
			population[curMember] = newPhenotype;
		}
	} else {
		//haven't finished running the initial population
		// so haven't mutated yet, so use this as the baseline
		population[curMember] = newPhenotype;
	}

	//statistics
	if (population[curMember].fitness>WorstFitness)
	{
		WorstFitness = population[curMember].fitness;
	}
	if (population[curMember].fitness<minfitness)
	{
		minfitness = population[curMember].fitness;
	}
	avgfitness+=population[curMember].fitness;
	runningMember=0;
}

/***********************************************************
* unsigned char ManageEvolution(void)
* 		Controls the evolutionary process and ensures that each
*		 step runs in the right order
************************************************************/
unsigned char ManageEvolution(void)
{
	unsigned char retVal=0;

	//loop through the phenotypes
	if (curMember<NUM_PHENOTYPES)
	{
		//is one running now?
		if (!runningMember)
		{
			//no, start the next one
			StartNextMember();
		}
		if (doneRunningMember)
		{
			//done running it, update the phenotype
			UpdateMember();
			//found a solution, show results
			if (population[curMember].fitness<=FITNESS_THRESHOLD)
			{
				sprintf(temp_str,"Found an acceptable solution: %d %d %d\n\r",
												population[curMember].motors.chars[0],
												population[curMember].motors.chars[1],
												population[curMember].motors.chars[2]);
				uart_send_str(temp_str,strlen(temp_str));
				retVal = 1;
			}
			//next member
			curMember++;
		}
	} else {
		//initial run?
		if (!hasInitialized)
		{
			//done initial run.
			hasInitialized=1;
		}

		//statistics
		displayavgfitness = avgfitness/NUM_PHENOTYPES;

		//display results
		if (WorstFitness<FITNESS_THRESHOLD)
		{
			sprintf(temp_str,"All members are better than fitness threshold.\n\r");
			uart_send_str(temp_str,strlen(temp_str));
			retVal = 1;
		}
		if ((displayavgfitness == minfitness) && (displayavgfitness == WorstFitness))
		{
			if (displayavgfitness<FITNESS_THRESHOLD)
			{
				sprintf(temp_str,"All members are equally better than fitness threshold.\n\r");
				uart_send_str(temp_str,strlen(temp_str));
				retVal = 1;
			}
		}
		//go to the next epoch
		StartNewEpoch();
	}
	//return whether or not a solution was found
	return retVal;
}
