/***********************************************************
UART communications code

William Westrick
Nicole Rodia
************************************************************/

#include <avr/io.h>			//for peripheral functionality
#include <avr/interrupt.h>	//for ISR functionality

#include "uart.h"			//for variables and prototypes

//buffer containing the received string before it is complete
volatile char rx_buffer[RX_BUFFER_SIZE];

//variables that maintain the position within the respective buffers
volatile unsigned int rx_pos;
volatile unsigned int tx_pos;

/***********************************************************
*  ISR(SIG_USART_RECV)
*		UART has received a byte, process it
************************************************************/
ISR(SIG_USART_RECV)
{
	unsigned int temp_char,i;
	temp_char=UDR0;			//store the byte temporarity

	if (temp_char=='\r') //return character?
	{
		if (rx_pos) //got at least 1 character that wasn't return?
		{
			rx_flag_received=1;	//received a full string + return character
						// don't increment the length of the received string
			rx_length_received = rx_pos;	//set the final received length
			//loop through the received data and put it into the buffer
			for (i=0;i<rx_length_received;i++)	
			{ rx_received[i] = rx_buffer[i]; }
			rx_received[rx_length_received+1]=0x00; //null character for string ops
			rx_pos = 0; //reset the position counter
			uart_send_str("\r\n",2);	//return a newline
		}
	} else {
		if (temp_char=='\b') //backspace?
		{
			//delete the last character from the display
			uart_send_str("\b \b",3);
			rx_pos--; //forget about the last entered character
		} else {
			if (temp_char!='\n') //don't process newlines
			{
				//add the character to the end of the buffer
				rx_buffer[rx_pos]=temp_char;
				uart_send_str(&rx_buffer[rx_pos],1);	//echo the character
				rx_pos++; //increment the character counter
			}
		}
	}
}

/***********************************************************
*  ISR(SIG_USART_DATA)
*		Transmit data register has at least one empty slot, fill it
************************************************************/
ISR(SIG_USART_DATA)
{
	UDR0 = tx_buffer[tx_pos]; //transmit the next byte
	tx_pos++;	//increment the position counter
	if (tx_pos == TX_BUFFER_SIZE) {tx_pos = 0;}	//loop around the buffer

	//is there more data left to transmit?
	//	if not, turn off the ISR
	tx_length_remaining--;
	if (!tx_length_remaining) { UCSR0B &= ~(1<<UDRIE0); }	
}

/***********************************************************
*  uart_init
*		Initialize UART communications
************************************************************/
void uart_init(void)
{
	//set the baud rate register
#if F_CPU < 2000000UL && defined(U2X)
  	UCSR0A = _BV(U2X);             /* improve baud rate error by using 2x clk */
  	UBRR0L = (F_CPU / (8UL * UART_BAUD)) - 1;
#else
  	UBRR0L = (F_CPU / (16UL * UART_BAUD)) - 1;
#endif
  	
	//set the UART control register
	UCSR0B = (1<<TXEN0) | (1<<RXEN0) | (1<<RXCIE0); /* tx/rx enable */
}

/***********************************************************
*  uart_send_str
*		send a string over UART with specified length
************************************************************/
unsigned char uart_send_str(volatile char *str, unsigned int length)
{
	//make sure input is relatively valid
	// if not, return with an error code
	if (!str) { return 0xff; } //null string
	if (!length) { return 0xfe; } // zero length
	//make sure that the buffer will not overflow, but allow for looping
	if ((length+tx_length_remaining)>TX_BUFFER_SIZE) { return 0xfd; }

	unsigned int i;
	unsigned int temp_tx_offset;
	unsigned int tx_offset;
	//figure out where the end of the currently transmitting data is
	temp_tx_offset = tx_pos+tx_length_remaining;
	//loop through the string and add it to the buffer
	for (i=0; i<length; i++)
	{
		//determine where the next character should go
		tx_offset = temp_tx_offset+i;
		//loop around if it exceeds the length of the buffer
		if (tx_offset>=TX_BUFFER_SIZE) { tx_offset-=TX_BUFFER_SIZE; }
		//put the character into the transmit buffer
		tx_buffer[tx_offset] = str[i];
	}
	//add the length of the new data to the length of the remaining data
	tx_length_remaining += length;
	UCSR0B |= (1<<UDRIE0); //enable the transmit ISR

	return 0;	//return with no error
}
